@@ -29,8 +29,11 @@ OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 #pragma once
 
 #include <initializer_list>
+#include <OSL/oslconfig.h>
 #include <OSL/oslversion.h>
 #include <OpenImageIO/fmath.h>
+
+
 OSL_NAMESPACE_ENTER
 
 
@@ -61,109 +64,109 @@ public:
 
     /// Default ctr leaves everything uninitialized
     ///
-    OIIO_CONSTEXPR14 Dual () { }
+    OSL_HOSTDEVICE OIIO_CONSTEXPR14 Dual () { }
 
     /// Construct a Dual from just a real value (derivs set to 0)
     ///
-    OIIO_CONSTEXPR14 Dual (const T &x) {
+    OSL_HOSTDEVICE OIIO_CONSTEXPR14 Dual (const T &x) {
         m_data[0] = x;
         for (int i = 1; i <= PARTIALS; ++i)
             m_data[i] = T(0.0);
     }
 
     /// Copy constructor from another Dual of same type and dimension.
-    OIIO_CONSTEXPR14 Dual (const Dual &x) {
+    OSL_HOSTDEVICE OIIO_CONSTEXPR14 Dual (const Dual &x) {
         for (int i = 0; i <= PARTIALS; ++i)
             m_data[i] = T(x.m_data[i]);
     }
 
-#if 1
     /// Copy constructor from another Dual of same dimension and different,
     /// but castable, data type.
     template<class F>
-    OIIO_CONSTEXPR14 Dual (const Dual<F,PARTIALS> &x) {
+    OSL_HOSTDEVICE OIIO_CONSTEXPR14 Dual (const Dual<F,PARTIALS> &x) {
         for (int i = 0; i <= PARTIALS; ++i)
             m_data[i] = T(x.elem(i));
     }
-#endif
 
     /// Construct a Dual from a real and infinitesimals.
     ///
-    constexpr Dual (const T &x, const T &dx) : m_data{ x, dx } {
+    OSL_HOSTDEVICE constexpr Dual (const T &x, const T &dx) : m_data{ x, dx } {
         static_assert(PARTIALS==1, "Wrong number of initializers");
     }
-    constexpr Dual (const T &x, const T &dx, const T &dy)
+    OSL_HOSTDEVICE constexpr Dual (const T &x, const T &dx, const T &dy)
         : m_data{ x, dx, dy }
     {
         static_assert(PARTIALS==2, "Wrong number of initializers");
     }
-    void set (const T &x, const T &dx) {
+    OSL_HOSTDEVICE void set (const T &x, const T &dx) {
         static_assert(PARTIALS==1, "Wrong number of initializers");
         m_data[0] = x;
         m_data[1] = dx;
     }
-    void set (const T &x, const T &dx, const T &dy) {
+    OSL_HOSTDEVICE void set (const T &x, const T &dx, const T &dy) {
         static_assert(PARTIALS==2, "Wrong number of initializers");
         m_data[0] = x;
         m_data[1] = dx;
         m_data[2] = dy;
     }
 
-    OIIO_CONSTEXPR14 Dual (std::initializer_list<T> vals) {
+    OSL_HOSTDEVICE OIIO_CONSTEXPR14 Dual (std::initializer_list<T> vals) {
+#if OIIO_CPLUSPLUS_VERSION >= 14
         static_assert (vals.size() == elements, "Wrong number of initializers");
+#endif
         for (int i = 0; i < elements; ++i)
             m_data[i] = vals.begin()[i];
     }
 
     /// Return the real value.
-    constexpr const T& val () const { return m_data[0]; }
-    T& val () { return m_data[0]; }
+    OSL_HOSTDEVICE constexpr const T& val () const { return m_data[0]; }
+    OSL_HOSTDEVICE T& val () { return m_data[0]; }
 
     /// Return the partial derivative with respect to x.
-    constexpr const T& dx () const { return m_data[1]; }
-    T& dx () { return m_data[1]; }
+    OSL_HOSTDEVICE constexpr const T& dx () const { return m_data[1]; }
+    OSL_HOSTDEVICE T& dx () { return m_data[1]; }
 
     /// Return the partial derivative with respect to y.
     /// Only valid if there are at least 2 partial dimensions.
-    constexpr const T& dy () const {
+    OSL_HOSTDEVICE constexpr const T& dy () const {
         static_assert(PARTIALS>=2, "Cannot call dy without at least 2 partials");
         return m_data[2];
     }
-    T& dy () {
+    OSL_HOSTDEVICE T& dy () {
         static_assert(PARTIALS>=2, "Cannot call dy without at least 2 partials");
         return m_data[2];
     }
 
     /// Return the partial derivative with respect to z.
     /// Only valid if there are at least 3 partial dimensions.
-    constexpr const T& dz () const {
+    OSL_HOSTDEVICE constexpr const T& dz () const {
         static_assert(PARTIALS>=3, "Cannot call dz without at least 3 partials");
         return m_data[3];
     }
-    T& dz () {
+    OSL_HOSTDEVICE T& dz () {
         static_assert(PARTIALS>=3, "Cannot call dz without at least 3 partials");
         return m_data[3];
     }
 
     /// Clear the derivatives; leave the value alone.
-    void clear_d () {
+    OSL_HOSTDEVICE void clear_d () {
         for (int i = 1; i <= PARTIALS; ++i)
             m_data[i] = T(0.0);
     }
 
     /// Return the i-th partial derivative.
-    constexpr const T& partial (int i) const {
+    OSL_HOSTDEVICE constexpr const T& partial (int i) const {
         return m_data[i+1];
     }
 
     /// Return a mutable reference to the i-th partial derivative.
     /// Use with caution!
-    T& partial (int i) {
+    OSL_HOSTDEVICE T& partial (int i) {
         return m_data[i+1];
     }
 
     /// Assignment of a real (the derivs are implicitly 0).
-    const Dual & operator= (const T &x) {
+    OSL_HOSTDEVICE const Dual & operator= (const T &x) {
         m_data[0] = x;
         for (int i = 1; i <= PARTIALS; ++i)
             m_data[i] = T(0);
@@ -172,8 +175,8 @@ public:
 
     /// Access like an array -- be careful! Element 0 is the main value,
     /// elements [1..PARTIALS] are the infinitessimals.
-    constexpr const T& elem (int i) const { return m_data[i]; }
-    T& elem (int i) { return m_data[i]; }
+    OSL_HOSTDEVICE constexpr const T& elem (int i) const { return m_data[i]; }
+    OSL_HOSTDEVICE T& elem (int i) { return m_data[i]; }
 
     /// Stream output.  Format as: "val[dx,dy,...]"
     ///
@@ -217,7 +220,7 @@ template<class T> using Dual2 = Dual<T,2>;
 /// Addition of duals.
 ///
 template<class T, int P>
-inline OIIO_CONSTEXPR14 Dual<T,P> operator+ (const Dual<T,P> &a, const Dual<T,P> &b)
+OSL_HOSTDEVICE inline OIIO_CONSTEXPR14 Dual<T,P> operator+ (const Dual<T,P> &a, const Dual<T,P> &b)
 {
     Dual<T,P> result = a;
     for (int i = 0; i <= P; ++i)
@@ -227,7 +230,7 @@ inline OIIO_CONSTEXPR14 Dual<T,P> operator+ (const Dual<T,P> &a, const Dual<T,P>
 
 
 template<class T, int P>
-inline OIIO_CONSTEXPR14 Dual<T,P> operator+ (const Dual<T,P> &a, const T &b)
+OSL_HOSTDEVICE inline OIIO_CONSTEXPR14 Dual<T,P> operator+ (const Dual<T,P> &a, const T &b)
 {
     Dual<T,P> result = a;
     result.val() += b;
@@ -236,7 +239,7 @@ inline OIIO_CONSTEXPR14 Dual<T,P> operator+ (const Dual<T,P> &a, const T &b)
 
 
 template<class T, int P>
-inline OIIO_CONSTEXPR14 Dual<T,P> operator+ (const T &a, const Dual<T,P> &b)
+OSL_HOSTDEVICE inline OIIO_CONSTEXPR14 Dual<T,P> operator+ (const T &a, const Dual<T,P> &b)
 {
     Dual<T,P> result = b;
     result.val() += a;
@@ -245,7 +248,7 @@ inline OIIO_CONSTEXPR14 Dual<T,P> operator+ (const T &a, const Dual<T,P> &b)
 
 
 template<class T, int P>
-inline Dual<T,P>& operator+= (Dual<T,P> &a, const Dual<T,P> &b)
+OSL_HOSTDEVICE inline Dual<T,P>& operator+= (Dual<T,P> &a, const Dual<T,P> &b)
 {
     for (int i = 0; i <= P; ++i)
         a.elem(i) += b.elem(i);
@@ -254,7 +257,7 @@ inline Dual<T,P>& operator+= (Dual<T,P> &a, const Dual<T,P> &b)
 
 
 template<class T, int P>
-inline Dual<T,P>& operator+= (Dual<T,P> &a, const T &b)
+OSL_HOSTDEVICE inline Dual<T,P>& operator+= (Dual<T,P> &a, const T &b)
 {
     a.val() += b;
     return a;
@@ -264,7 +267,7 @@ inline Dual<T,P>& operator+= (Dual<T,P> &a, const T &b)
 /// Subtraction of duals.
 ///
 template<class T, int P>
-inline OIIO_CONSTEXPR14 Dual<T,P> operator- (const Dual<T,P> &a, const Dual<T,P> &b)
+OSL_HOSTDEVICE inline OIIO_CONSTEXPR14 Dual<T,P> operator- (const Dual<T,P> &a, const Dual<T,P> &b)
 {
     Dual<T,P> result;
     for (int i = 0; i <= P; ++i)
@@ -274,7 +277,7 @@ inline OIIO_CONSTEXPR14 Dual<T,P> operator- (const Dual<T,P> &a, const Dual<T,P>
 
 
 template<class T, int P>
-inline OIIO_CONSTEXPR14 Dual<T,P> operator- (const Dual<T,P> &a, const T &b)
+OSL_HOSTDEVICE inline OIIO_CONSTEXPR14 Dual<T,P> operator- (const Dual<T,P> &a, const T &b)
 {
     Dual<T,P> result = a;
     result.val() -= b;
@@ -283,7 +286,7 @@ inline OIIO_CONSTEXPR14 Dual<T,P> operator- (const Dual<T,P> &a, const T &b)
 
 
 template<class T, int P>
-inline OIIO_CONSTEXPR14 Dual<T,P> operator- (const T &a, const Dual<T,P> &b)
+OSL_HOSTDEVICE inline OIIO_CONSTEXPR14 Dual<T,P> operator- (const T &a, const Dual<T,P> &b)
 {
     Dual<T,P> result;
     result.val() = a - b.val();
@@ -294,7 +297,7 @@ inline OIIO_CONSTEXPR14 Dual<T,P> operator- (const T &a, const Dual<T,P> &b)
 
 
 template<class T, int P>
-inline Dual<T,P>& operator-= (Dual<T,P> &a, const Dual<T,P> &b)
+OSL_HOSTDEVICE inline Dual<T,P>& operator-= (Dual<T,P> &a, const Dual<T,P> &b)
 {
     for (int i = 0; i <= P; ++i)
         a.elem(i) -= b.elem(i);
@@ -303,7 +306,7 @@ inline Dual<T,P>& operator-= (Dual<T,P> &a, const Dual<T,P> &b)
 
 
 template<class T, int P>
-inline Dual<T,P>& operator-= (Dual<T,P> &a, const T &b)
+OSL_HOSTDEVICE inline Dual<T,P>& operator-= (Dual<T,P> &a, const T &b)
 {
     a.val() -= b.val();
     return a;
@@ -314,7 +317,7 @@ inline Dual<T,P>& operator-= (Dual<T,P> &a, const T &b)
 /// Negation of duals.
 ///
 template<class T, int P>
-inline OIIO_CONSTEXPR14 Dual<T,P> operator- (const Dual<T,P> &a)
+OSL_HOSTDEVICE inline OIIO_CONSTEXPR14 Dual<T,P> operator- (const Dual<T,P> &a)
 {
     Dual<T,P> result;
     for (int i = 0; i <= P; ++i)
@@ -326,7 +329,7 @@ inline OIIO_CONSTEXPR14 Dual<T,P> operator- (const Dual<T,P> &a)
 /// Multiplication of duals. This will work for any Dual<T>*Dual<S>
 /// where T*S is meaningful.
 template<class T, int P, class S>
-inline OIIO_CONSTEXPR14 auto
+OSL_HOSTDEVICE inline OIIO_CONSTEXPR14 auto
 operator* (const Dual<T,P> &a, const Dual<S,P> &b) -> Dual<decltype(a.elem(0)*b.elem(0)),P>
 {
     // Use the chain rule
@@ -342,7 +345,7 @@ operator* (const Dual<T,P> &a, const Dual<S,P> &b) -> Dual<decltype(a.elem(0)*b.
 /// Dual<T> * S where T*S is meaningful.
 template<class T, int P, class S,
          DUAL_REQUIRES(is_Dual<S>::value == false)>
-inline OIIO_CONSTEXPR14 auto
+OSL_HOSTDEVICE inline OIIO_CONSTEXPR14 auto
 operator* (const Dual<T,P> &a, const S &b) -> Dual<decltype(a.elem(0)*b),P>
 {
     Dual<decltype(a.elem(0)*b),P> result;
@@ -356,7 +359,7 @@ operator* (const Dual<T,P> &a, const S &b) -> Dual<decltype(a.elem(0)*b),P>
 ///
 template<class T, int P, class S,
          DUAL_REQUIRES(is_Dual<S>::value == false)>
-inline const Dual<T,P>& operator*= (Dual<T,P> &a, const S &b)
+OSL_HOSTDEVICE inline const Dual<T,P>& operator*= (Dual<T,P> &a, const S &b)
 {
     for (int i = 0; i <= P; ++i)
         a.elem(i) *= b;
@@ -369,7 +372,7 @@ inline const Dual<T,P>& operator*= (Dual<T,P> &a, const S &b)
 /// Dual<T> * S where T*S is meaningful.
 template<class T, int P, class S,
          DUAL_REQUIRES(is_Dual<S>::value == false)>
-inline OIIO_CONSTEXPR14 auto
+OSL_HOSTDEVICE inline OIIO_CONSTEXPR14 auto
 operator* (const S &b, const Dual<T,P> &a) -> Dual<decltype(a.elem(0)*b),P>
 {
     return a*b;
@@ -380,10 +383,10 @@ operator* (const S &b, const Dual<T,P> &a) -> Dual<decltype(a.elem(0)*b),P>
 /// Division of duals.
 ///
 template<class T, int P>
-inline OIIO_CONSTEXPR14 Dual<T,P> operator/ (const Dual<T,P> &a, const Dual<T,P> &b)
+OSL_HOSTDEVICE inline OIIO_CONSTEXPR14 Dual<T,P> operator/ (const Dual<T,P> &a, const Dual<T,P> &b)
 {
-    T bvalinv = 1.0f / b.val();
-    T aval_bval = a.val() * bvalinv;
+    T bvalinv = T(1) / b.val();
+    T aval_bval = a.val() / b.val();
     Dual<T,P> result;
     result.val() = aval_bval;
     for (int i = 1; i <= P; ++i)
@@ -395,20 +398,25 @@ inline OIIO_CONSTEXPR14 Dual<T,P> operator/ (const Dual<T,P> &a, const Dual<T,P>
 /// Division of dual by scalar.
 ///
 template<class T, int P>
-inline OIIO_CONSTEXPR14 Dual<T,P> operator/ (const Dual<T,P> &a, const T &b)
+OSL_HOSTDEVICE inline OIIO_CONSTEXPR14 Dual<T,P> operator/ (const Dual<T,P> &a, const T &b)
 {
-    T binv = 1.0f / b;
-    return a * binv;
+    T bvalinv = T(1) / b;
+    T aval_bval = a.val() / b;
+    Dual<T,P> result;
+    result.val() = aval_bval;
+    for (int i = 1; i <= P; ++i)
+        result.elem(i) = bvalinv * a.elem(i);
+    return result;
 }
 
 
 /// Division of scalar by dual.
 ///
 template<class T, int P>
-inline OIIO_CONSTEXPR14 Dual<T,P> operator/ (const T &aval, const Dual<T,P> &b)
+OSL_HOSTDEVICE inline OIIO_CONSTEXPR14 Dual<T,P> operator/ (const T &aval, const Dual<T,P> &b)
 {
-    T bvalinv = 1.0f / b.val();
-    T aval_bval = aval * bvalinv;
+    T bvalinv = T(1) / b.val();
+    T aval_bval = aval / b.val();
     Dual<T,P> result;
     result.val() = aval_bval;
     for (int i = 1; i <= P; ++i)
@@ -420,75 +428,75 @@ inline OIIO_CONSTEXPR14 Dual<T,P> operator/ (const T &aval, const Dual<T,P> &b)
 
 
 template<class T, int P>
-inline constexpr bool operator< (const Dual<T,P> &a, const Dual<T,P> &b) {
+OSL_HOSTDEVICE inline constexpr bool operator< (const Dual<T,P> &a, const Dual<T,P> &b) {
     return a.val() < b.val();
 }
 
 template<class T, int P>
-inline constexpr bool operator< (const Dual<T,P> &a, const T &b) {
+OSL_HOSTDEVICE inline constexpr bool operator< (const Dual<T,P> &a, const T &b) {
     return a.val() < b;
 }
 
 template<class T, int P>
-inline constexpr bool operator> (const Dual<T,P> &a, const Dual<T,P> &b) {
+OSL_HOSTDEVICE inline constexpr bool operator> (const Dual<T,P> &a, const Dual<T,P> &b) {
     return a.val() > b.val();
 }
 
 template<class T, int P>
-inline constexpr bool operator> (const Dual<T,P> &a, const T &b) {
+OSL_HOSTDEVICE inline constexpr bool operator> (const Dual<T,P> &a, const T &b) {
     return a.val() > b;
 }
 
 template<class T, int P>
-inline constexpr bool operator<= (const Dual<T,P> &a, const Dual<T,P> &b) {
+OSL_HOSTDEVICE inline constexpr bool operator<= (const Dual<T,P> &a, const Dual<T,P> &b) {
     return a.val() <= b.val();
 }
 
 template<class T, int P>
-inline constexpr bool operator<= (const Dual<T,P> &a, const T &b) {
+OSL_HOSTDEVICE inline constexpr bool operator<= (const Dual<T,P> &a, const T &b) {
     return a.val() <= b;
 }
 
 template<class T, int P>
-inline constexpr bool operator>= (const Dual<T,P> &a, const Dual<T,P> &b) {
+OSL_HOSTDEVICE inline constexpr bool operator>= (const Dual<T,P> &a, const Dual<T,P> &b) {
     return a.val() >= b.val();
 }
 
 template<class T, int P>
-inline constexpr bool operator>= (const Dual<T,P> &a, const T &b) {
+OSL_HOSTDEVICE inline constexpr bool operator>= (const Dual<T,P> &a, const T &b) {
     return a.val() >= b;
 }
 
 
 
 // Eliminate the derivatives of a number, works for scalar as well as Dual.
-template<class T> inline constexpr const T&
+template<class T> OSL_HOSTDEVICE inline constexpr const T&
 removeDerivatives (const T &x) { return x; }
 
-template<class T, int P> inline constexpr const T&
+template<class T, int P> OSL_HOSTDEVICE inline constexpr const T&
 removeDerivatives (const Dual<T,P> &x) { return x.val(); }
 
 
 // Get the x derivative (or 0 for a non-Dual)
-template<class T> inline constexpr const T&
+template<class T> OSL_HOSTDEVICE inline constexpr const T&
 getXDerivative (const T &x) { return T(0); }
 
-template<class T, int P> inline constexpr const T&
+template<class T, int P> OSL_HOSTDEVICE inline constexpr const T&
 getXDerivative (const Dual<T,P> &x) { return x.partial(0); }
 
 
 // Get the y derivative (or 0 for a non-Dual)
-template<class T> inline constexpr const T&
+template<class T> OSL_HOSTDEVICE inline constexpr const T&
 getYDerivative (const T &x) { return T(0); }
 
-template<class T, int P> inline constexpr const T&
+template<class T, int P> OSL_HOSTDEVICE inline constexpr const T&
 getYDerivative (const Dual<T,P> &x) { return x.partial(1); }
 
 
 // Simple templated "copy" function
-template<class T> inline void
+template<class T> OSL_HOSTDEVICE inline void
 assignment (T &a, T &b) { a = b; }
-template<class T, int P> inline void
+template<class T, int P> OSL_HOSTDEVICE inline void
 assignment (T &a, Dual<T,P> &b) { a = b.val(); }
 
 // Templated value equality. For scalars, it's the same as regular ==.
@@ -497,22 +505,22 @@ assignment (T &a, Dual<T,P> &b) { a = b.val(); }
 // return if just their value is equal or if the whole struct (including
 // derivs) are equal.
 template<class T>
-inline constexpr bool equalVal (const T &x, const T &y) {
+OSL_HOSTDEVICE inline constexpr bool equalVal (const T &x, const T &y) {
     return x == y;
 }
 
 template<class T, int P>
-inline constexpr bool equalVal (const Dual<T,P> &x, const Dual<T,P> &y) {
+OSL_HOSTDEVICE inline constexpr bool equalVal (const Dual<T,P> &x, const Dual<T,P> &y) {
     return x.val() == y.val();
 }
 
 template<class T, int P>
-inline constexpr bool equalVal (const Dual<T,P> &x, const T &y) {
+OSL_HOSTDEVICE inline constexpr bool equalVal (const Dual<T,P> &x, const T &y) {
     return x.val() == y;
 }
 
 template<class T, int P>
-inline constexpr bool equalVal (const T &x, const Dual<T,P> &y) {
+OSL_HOSTDEVICE inline constexpr bool equalVal (const T &x, const Dual<T,P> &y) {
     return x == y.val();
 }
 
@@ -523,7 +531,7 @@ inline constexpr bool equalVal (const T &x, const Dual<T,P> &y) {
 // function f(scalar), then the dual function F(<u,u'>) is defined as:
 //    F(<u,u'>) = < f(u), f'(u)*u' >
 template<class T, int P>
-inline OIIO_CONSTEXPR14 Dual<T,P>
+OSL_HOSTDEVICE inline OIIO_CONSTEXPR14 Dual<T,P>
 dualfunc (const Dual<T,P>& u, const T& f_val, const T& df_val)
 {
     Dual<T,P> result;
@@ -539,7 +547,7 @@ dualfunc (const Dual<T,P>& u, const T& f_val, const T& df_val)
 //   F(<u,u'>, <v,v'>) = < f(u,v), dfdu(u,v)u' + dfdv(u,v)v' >
 // (from http://en.wikipedia.org/wiki/Automatic_differentiation)
 template<class T, int P>
-inline OIIO_CONSTEXPR14 Dual<T,P>
+OSL_HOSTDEVICE inline OIIO_CONSTEXPR14 Dual<T,P>
 dualfunc (const Dual<T,P>& u, const Dual<T,P>& v,
           const T& f_val, const T& dfdu_val, const T& dfdv_val)
 {
@@ -553,7 +561,7 @@ dualfunc (const Dual<T,P>& u, const Dual<T,P>& v,
 
 // Helper for construction the result of a Dual function of three variables.
 template<class T, int P>
-inline OIIO_CONSTEXPR14 Dual<T,P>
+OSL_HOSTDEVICE inline OIIO_CONSTEXPR14 Dual<T,P>
 dualfunc (const Dual<T,P>& u, const Dual<T,P>& v, const Dual<T,P>& w,
           const T& f_val, const T& dfdu_val, const T& dfdv_val, const T& dfdw_val)
 {
@@ -568,7 +576,7 @@ dualfunc (const Dual<T,P>& u, const Dual<T,P>& v, const Dual<T,P>& w,
 
 // f(x) = cos(x), f'(x) = -sin(x)
 template<class T, int P>
-inline Dual<T,P> cos (const Dual<T,P> &a)
+OSL_HOSTDEVICE inline Dual<T,P> cos (const Dual<T,P> &a)
 {
     T sina, cosa;
     OIIO::sincos(a.val(), &sina, &cosa);
@@ -576,7 +584,7 @@ inline Dual<T,P> cos (const Dual<T,P> &a)
 }
 
 template<class T, int P>
-inline Dual<T,P> fast_cos(const Dual<T,P> &a)
+OSL_HOSTDEVICE inline Dual<T,P> fast_cos(const Dual<T,P> &a)
 {
     T sina, cosa;
     OIIO::fast_sincos (a.val(), &sina, &cosa);
@@ -585,7 +593,7 @@ inline Dual<T,P> fast_cos(const Dual<T,P> &a)
 
 // f(x) = sin(x),  f'(x) = cos(x)
 template<class T, int P>
-inline Dual<T,P> sin (const Dual<T,P> &a)
+OSL_HOSTDEVICE inline Dual<T,P> sin (const Dual<T,P> &a)
 {
     T sina, cosa;
     OIIO::sincos(a.val(), &sina, &cosa);
@@ -593,7 +601,7 @@ inline Dual<T,P> sin (const Dual<T,P> &a)
 }
 
 template<class T, int P>
-inline Dual<T,P> fast_sin(const Dual<T,P> &a)
+OSL_HOSTDEVICE inline Dual<T,P> fast_sin(const Dual<T,P> &a)
 {
     T sina, cosa;
     OIIO::fast_sincos (a.val(), &sina, &cosa);
@@ -601,7 +609,7 @@ inline Dual<T,P> fast_sin(const Dual<T,P> &a)
 }
 
 template<class T, int P>
-inline void sincos(const Dual<T,P> &a, Dual<T,P> *sine, Dual<T,P> *cosine)
+OSL_HOSTDEVICE inline void sincos(const Dual<T,P> &a, Dual<T,P> *sine, Dual<T,P> *cosine)
 {
 	T sina, cosa;
 	OIIO::sincos(a.val(), &sina, &cosa);
@@ -610,7 +618,7 @@ inline void sincos(const Dual<T,P> &a, Dual<T,P> *sine, Dual<T,P> *cosine)
 }
 
 template<class T, int P>
-inline void fast_sincos(const Dual<T,P> &a, Dual<T,P> *sine, Dual<T,P> *cosine)
+OSL_HOSTDEVICE inline void fast_sincos(const Dual<T,P> &a, Dual<T,P> *sine, Dual<T,P> *cosine)
 {
 	T sina, cosa;
 	OIIO::fast_sincos(a.val(), &sina, &cosa);
@@ -620,7 +628,7 @@ inline void fast_sincos(const Dual<T,P> &a, Dual<T,P> *sine, Dual<T,P> *cosine)
 
 // f(x) = tan(x), f'(x) = sec^2(x)
 template<class T, int P>
-inline Dual<T,P> tan (const Dual<T,P> &a)
+OSL_HOSTDEVICE inline Dual<T,P> tan (const Dual<T,P> &a)
 {
     T tana  = std::tan (a.val());
     T cosa  = std::cos (a.val());
@@ -629,7 +637,7 @@ inline Dual<T,P> tan (const Dual<T,P> &a)
 }
 
 template<class T, int P>
-inline Dual<T,P> fast_tan(const Dual<T,P> &a)
+OSL_HOSTDEVICE inline Dual<T,P> fast_tan(const Dual<T,P> &a)
 {
     T tana  = OIIO::fast_tan (a.val());
     T cosa  = OIIO::fast_cos (a.val());
@@ -639,7 +647,7 @@ inline Dual<T,P> fast_tan(const Dual<T,P> &a)
 
 // f(x) = cosh(x), f'(x) = sinh(x)
 template<class T, int P>
-inline Dual<T,P> cosh (const Dual<T,P> &a)
+OSL_HOSTDEVICE inline Dual<T,P> cosh (const Dual<T,P> &a)
 {
     T f = std::cosh(a.val());
     T df = std::sinh(a.val());
@@ -647,7 +655,7 @@ inline Dual<T,P> cosh (const Dual<T,P> &a)
 }
 
 template<class T, int P>
-inline Dual<T,P> fast_cosh(const Dual<T,P> &a)
+OSL_HOSTDEVICE inline Dual<T,P> fast_cosh(const Dual<T,P> &a)
 {
     T f = OIIO::fast_cosh(a.val());
     T df = OIIO::fast_sinh(a.val());
@@ -657,7 +665,7 @@ inline Dual<T,P> fast_cosh(const Dual<T,P> &a)
 
 // f(x) = sinh(x), f'(x) = cosh(x)
 template<class T, int P>
-inline Dual<T,P> sinh (const Dual<T,P> &a)
+OSL_HOSTDEVICE inline Dual<T,P> sinh (const Dual<T,P> &a)
 {
     T f = std::sinh(a.val());
     T df = std::cosh(a.val());
@@ -665,7 +673,7 @@ inline Dual<T,P> sinh (const Dual<T,P> &a)
 }
 
 template<class T, int P>
-inline Dual<T,P> fast_sinh(const Dual<T,P> &a)
+OSL_HOSTDEVICE inline Dual<T,P> fast_sinh(const Dual<T,P> &a)
 {
     T f = OIIO::fast_sinh(a.val());
     T df = OIIO::fast_cosh(a.val());
@@ -674,7 +682,7 @@ inline Dual<T,P> fast_sinh(const Dual<T,P> &a)
 
 // f(x) = tanh(x), f'(x) = sech^2(x)
 template<class T, int P>
-inline Dual<T,P> tanh (const Dual<T,P> &a)
+OSL_HOSTDEVICE inline Dual<T,P> tanh (const Dual<T,P> &a)
 {
     T tanha = std::tanh(a.val());
     T cosha = std::cosh(a.val());
@@ -683,7 +691,7 @@ inline Dual<T,P> tanh (const Dual<T,P> &a)
 }
 
 template<class T, int P>
-inline Dual<T,P> fast_tanh(const Dual<T,P> &a)
+OSL_HOSTDEVICE inline Dual<T,P> fast_tanh(const Dual<T,P> &a)
 {
     T tanha = OIIO::fast_tanh(a.val());
     T cosha = OIIO::fast_cosh(a.val());
@@ -693,7 +701,7 @@ inline Dual<T,P> fast_tanh(const Dual<T,P> &a)
 
 // f(x) = acos(x), f'(x) = -1/(sqrt(1 - x^2))
 template<class T, int P>
-inline Dual<T,P> safe_acos (const Dual<T,P> &a)
+OSL_HOSTDEVICE inline Dual<T,P> safe_acos (const Dual<T,P> &a)
 {
     if (a.val() >= T(1))
         return Dual<T,P> (T(0));
@@ -705,7 +713,7 @@ inline Dual<T,P> safe_acos (const Dual<T,P> &a)
 }
 
 template<class T, int P>
-inline Dual<T,P> fast_acos(const Dual<T,P> &a)
+OSL_HOSTDEVICE inline Dual<T,P> fast_acos(const Dual<T,P> &a)
 {
     T f = OIIO::fast_acos(a.val());
     T df = fabsf(a.val()) < 1.0f ? -1.0f / sqrtf(1.0f - a.val() * a.val()) : 0.0f;
@@ -714,7 +722,7 @@ inline Dual<T,P> fast_acos(const Dual<T,P> &a)
 
 // f(x) = asin(x), f'(x) = 1/(sqrt(1 - x^2))
 template<class T, int P>
-inline Dual<T,P> safe_asin (const Dual<T,P> &a)
+OSL_HOSTDEVICE inline Dual<T,P> safe_asin (const Dual<T,P> &a)
 {
     if (a.val() >= T(1))
         return Dual<T,P> (T(M_PI/2));
@@ -726,7 +734,7 @@ inline Dual<T,P> safe_asin (const Dual<T,P> &a)
 }
 
 template<class T, int P>
-inline Dual<T,P> fast_asin(const Dual<T,P> &a)
+OSL_HOSTDEVICE inline Dual<T,P> fast_asin(const Dual<T,P> &a)
 {
     T f = OIIO::fast_asin(a.val());
     T df = fabsf(a.val()) < 1.0f ? 1.0f / sqrtf(1.0f - a.val() * a.val()) : 0.0f;
@@ -736,7 +744,7 @@ inline Dual<T,P> fast_asin(const Dual<T,P> &a)
 
 // f(x) = atan(x), f'(x) = 1/(1 + x^2)
 template<class T, int P>
-inline Dual<T,P> atan (const Dual<T,P> &a)
+OSL_HOSTDEVICE inline Dual<T,P> atan (const Dual<T,P> &a)
 {
     T f = std::atan (a.val());
     T df = T(1) / (T(1) + a.val()*a.val());
@@ -744,7 +752,7 @@ inline Dual<T,P> atan (const Dual<T,P> &a)
 }
 
 template<class T, int P>
-inline Dual<T,P> fast_atan(const Dual<T,P> &a)
+OSL_HOSTDEVICE inline Dual<T,P> fast_atan(const Dual<T,P> &a)
 {
     T f = OIIO::fast_atan(a.val());
     T df = 1.0f / (1.0f + a.val() * a.val());
@@ -756,7 +764,7 @@ inline Dual<T,P> fast_atan(const Dual<T,P> &a)
 // reference:  http://en.wikipedia.org/wiki/Atan2
 // (above link has other formulations)
 template<class T, int P>
-inline Dual<T,P> atan2 (const Dual<T,P> &y, const Dual<T,P> &x)
+OSL_HOSTDEVICE inline Dual<T,P> atan2 (const Dual<T,P> &y, const Dual<T,P> &x)
 {
     T atan2xy = std::atan2 (y.val(), x.val());
     T denom = (x.val() == T(0) && y.val() == T(0)) ? T(0) : T(1) / (x.val()*x.val() + y.val()*y.val());
@@ -764,7 +772,7 @@ inline Dual<T,P> atan2 (const Dual<T,P> &y, const Dual<T,P> &x)
 }
 
 template<class T, int P>
-inline Dual<T,P> fast_atan2(const Dual<T,P> &y, const Dual<T,P> &x)
+OSL_HOSTDEVICE inline Dual<T,P> fast_atan2(const Dual<T,P> &y, const Dual<T,P> &x)
 {
     T atan2xy = OIIO::fast_atan2(y.val(), x.val());
     T denom = (x.val() == 0 && y.val() == 0) ? 0.0f : 1.0f / (x.val() * x.val() + y.val() * y.val());
@@ -779,7 +787,7 @@ inline Dual<T,P> fast_atan2(const Dual<T,P> &y, const Dual<T,P> &x)
 //   (from http://en.wikipedia.org/wiki/Automatic_differentiation)
 // so, pow(u,v) = < u^v, vu^(v-1) u' + log(u)u^v v' >
 template<class T, int P>
-inline Dual<T,P> safe_pow (const Dual<T,P> &u, const Dual<T,P> &v)
+OSL_HOSTDEVICE inline Dual<T,P> safe_pow (const Dual<T,P> &u, const Dual<T,P> &v)
 {
     // NOTE: this function won't return exactly the same as pow(x,y) because we
     // use the identity u^v=u * u^(v-1) which does not hold in all cases for our
@@ -791,7 +799,7 @@ inline Dual<T,P> safe_pow (const Dual<T,P> &u, const Dual<T,P> &v)
 }
 
 template<class T, int P>
-inline Dual<T,P> fast_safe_pow(const Dual<T,P> &u, const Dual<T,P> &v)
+OSL_HOSTDEVICE inline Dual<T,P> fast_safe_pow(const Dual<T,P> &u, const Dual<T,P> &v)
 {
     // NOTE: same issue as above (fast_safe_pow does even more clamping)
     T powuvm1 = OIIO::fast_safe_pow (u.val(), v.val() - 1.0f);
@@ -803,7 +811,7 @@ inline Dual<T,P> fast_safe_pow(const Dual<T,P> &u, const Dual<T,P> &v)
 // f(x) = log(a), f'(x) = 1/x
 // (log base e)
 template<class T, int P>
-inline Dual<T,P> safe_log (const Dual<T,P> &a)
+OSL_HOSTDEVICE inline Dual<T,P> safe_log (const Dual<T,P> &a)
 {
     T f = safe_log(a.val());
     T df = a.val() < std::numeric_limits<T>::min() ? T(0) : T(1) / a.val();
@@ -811,7 +819,7 @@ inline Dual<T,P> safe_log (const Dual<T,P> &a)
 }
 
 template<class T, int P>
-inline Dual<T,P> fast_log(const Dual<T,P> &a)
+OSL_HOSTDEVICE inline Dual<T,P> fast_log(const Dual<T,P> &a)
 {
     T f = OIIO::fast_log(a.val());
     T df = a.val() < std::numeric_limits<float>::min() ? 0.0f : 1.0f / a.val();
@@ -821,7 +829,7 @@ inline Dual<T,P> fast_log(const Dual<T,P> &a)
 // f(x) = log2(x), f'(x) = 1/(x*log2)
 // (log base 2)
 template<class T, int P>
-inline Dual<T,P> safe_log2 (const Dual<T,P> &a)
+OSL_HOSTDEVICE inline Dual<T,P> safe_log2 (const Dual<T,P> &a)
 {
     T f = safe_log2(a.val());
     T df = a.val() < std::numeric_limits<T>::min() ? T(0) : T(1) / (a.val() * T(M_LN2));
@@ -829,7 +837,7 @@ inline Dual<T,P> safe_log2 (const Dual<T,P> &a)
 }
 
 template<class T, int P>
-inline Dual<T,P> fast_log2(const Dual<T,P> &a)
+OSL_HOSTDEVICE inline Dual<T,P> fast_log2(const Dual<T,P> &a)
 {
     T f = OIIO::fast_log2(a.val());
     T aln2 = a.val() * float(M_LN2);
@@ -840,7 +848,7 @@ inline Dual<T,P> fast_log2(const Dual<T,P> &a)
 // f(x) = log10(x), f'(x) = 1/(x*log10)
 // (log base 10)
 template<class T, int P>
-inline Dual<T,P> safe_log10 (const Dual<T,P> &a)
+OSL_HOSTDEVICE inline Dual<T,P> safe_log10 (const Dual<T,P> &a)
 {
     T f = safe_log10(a.val());
     T df = a.val() < std::numeric_limits<T>::min() ? T(0) : T(1) / (a.val() * T(M_LN10));
@@ -848,7 +856,7 @@ inline Dual<T,P> safe_log10 (const Dual<T,P> &a)
 }
 
 template<class T, int P>
-inline Dual<T,P> fast_log10(const Dual<T,P> &a)
+OSL_HOSTDEVICE inline Dual<T,P> fast_log10(const Dual<T,P> &a)
 {
     T f  = OIIO::fast_log10(a.val());
     T aln10 = a.val() * float(M_LN10);
@@ -858,14 +866,14 @@ inline Dual<T,P> fast_log10(const Dual<T,P> &a)
 
 // f(x) = e^x, f'(x) = e^x
 template<class T, int P>
-inline Dual<T,P> exp (const Dual<T,P> &a)
+OSL_HOSTDEVICE inline Dual<T,P> exp (const Dual<T,P> &a)
 {
     T f = std::exp(a.val());
     return dualfunc (a, f, f);
 }
 
 template<class T, int P>
-inline Dual<T,P> fast_exp(const Dual<T,P> &a)
+OSL_HOSTDEVICE inline Dual<T,P> fast_exp(const Dual<T,P> &a)
 {
     T f = OIIO::fast_exp(a.val());
     return dualfunc (a, f, f);
@@ -873,7 +881,7 @@ inline Dual<T,P> fast_exp(const Dual<T,P> &a)
 
 // f(x) = 2^x, f'(x) = (2^x)*log(2)
 template<class T, int P>
-inline Dual<T,P> exp2 (const Dual<T,P> &a)
+OSL_HOSTDEVICE inline Dual<T,P> exp2 (const Dual<T,P> &a)
 {
     using std::exp2;
     T f = exp2(a.val());
@@ -881,7 +889,7 @@ inline Dual<T,P> exp2 (const Dual<T,P> &a)
 }
 
 template<class T, int P>
-inline Dual<T,P> fast_exp2(const Dual<T,P> &a)
+OSL_HOSTDEVICE inline Dual<T,P> fast_exp2(const Dual<T,P> &a)
 {
     T f = OIIO::fast_exp2(float(a.val()));
     return dualfunc (a, f, f*T(M_LN2));
@@ -890,7 +898,7 @@ inline Dual<T,P> fast_exp2(const Dual<T,P> &a)
 
 // f(x) = e^x - 1, f'(x) = e^x
 template<class T, int P>
-inline Dual<T,P> expm1 (const Dual<T,P> &a)
+OSL_HOSTDEVICE inline Dual<T,P> expm1 (const Dual<T,P> &a)
 {
     using std::expm1;
     T f  = expm1(a.val());
@@ -899,7 +907,7 @@ inline Dual<T,P> expm1 (const Dual<T,P> &a)
 }
 
 template<class T, int P>
-inline Dual<T,P> fast_expm1(const Dual<T,P> &a)
+OSL_HOSTDEVICE inline Dual<T,P> fast_expm1(const Dual<T,P> &a)
 {
     T f  = OIIO::fast_expm1(a.val());
     T df = OIIO::fast_exp  (a.val());
@@ -908,7 +916,7 @@ inline Dual<T,P> fast_expm1(const Dual<T,P> &a)
 
 // f(x) = erf(x), f'(x) = (2e^(-x^2))/sqrt(pi)
 template<class T, int P>
-inline Dual<T,P> erf (const Dual<T,P> &a)
+OSL_HOSTDEVICE inline Dual<T,P> erf (const Dual<T,P> &a)
 {
     using std::erf;
     T f = erf (a.val());
@@ -918,7 +926,7 @@ inline Dual<T,P> erf (const Dual<T,P> &a)
 }
 
 template<class T, int P>
-inline Dual<T,P> fast_erf(const Dual<T,P> &a)
+OSL_HOSTDEVICE inline Dual<T,P> fast_erf(const Dual<T,P> &a)
 {
     T f = OIIO::fast_erf (float(a.val())); // float version!
     const T two_over_sqrt_pi = 1.128379167095512573896158903f;
@@ -928,7 +936,7 @@ inline Dual<T,P> fast_erf(const Dual<T,P> &a)
 
 // f(x) = erfc(x), f'(x) = -(2e^(-x^2))/sqrt(pi)
 template<class T, int P>
-inline Dual<T,P> erfc (const Dual<T,P> &a)
+OSL_HOSTDEVICE inline Dual<T,P> erfc (const Dual<T,P> &a)
 {
     using std::erfc;
     T f = erfc (a.val()); // float version!
@@ -938,7 +946,7 @@ inline Dual<T,P> erfc (const Dual<T,P> &a)
 }
 
 template<class T, int P>
-inline Dual<T,P> fast_erfc(const Dual<T,P> &a)
+OSL_HOSTDEVICE inline Dual<T,P> fast_erfc(const Dual<T,P> &a)
 {
     T f = OIIO::fast_erfc (float(a.val())); // float version!
     const T two_over_sqrt_pi = -1.128379167095512573896158903f;
@@ -949,7 +957,7 @@ inline Dual<T,P> fast_erfc(const Dual<T,P> &a)
 
 // f(x) = sqrt(x), f'(x) = 1/(2*sqrt(x))
 template<class T, int P>
-inline OIIO_CONSTEXPR14 Dual<T,P> sqrt (const Dual<T,P> &a)
+OSL_HOSTDEVICE inline OIIO_CONSTEXPR14 Dual<T,P> sqrt (const Dual<T,P> &a)
 {
     if (a.val() <= T(0))
         return Dual<T,P> (T(0));
@@ -960,7 +968,7 @@ inline OIIO_CONSTEXPR14 Dual<T,P> sqrt (const Dual<T,P> &a)
 
 // f(x) = 1/sqrt(x), f'(x) = -1/(2*x^(3/2))
 template<class T, int P>
-inline Dual<T,P> inversesqrt (const Dual<T,P> &a)
+OSL_HOSTDEVICE inline Dual<T,P> inversesqrt (const Dual<T,P> &a)
 {
     // do we want to print an error message?
     if (a.val() <= T(0))
@@ -972,21 +980,21 @@ inline Dual<T,P> inversesqrt (const Dual<T,P> &a)
 
 // (fx) = x*(1-a) + y*a, f'(x) = (1-a)x' + (y - x)*a' + a*y'
 template<class T, int P>
-inline Dual<T,P> mix (const Dual<T,P> &x, const Dual<T,P> &y, const Dual<T,P> &a)
+OSL_HOSTDEVICE inline Dual<T,P> mix (const Dual<T,P> &x, const Dual<T,P> &y, const Dual<T,P> &a)
 {
     T mixval = x.val()*(T(1)-a.val()) + y.val()*a.val();
     return dualfunc (x, y, a, mixval, T(1)-a.val(), a.val(), y.val() - x.val());
 }
 
 template<class T, int P>
-inline Dual<T,P> fabs (const Dual<T,P> &x)
+OSL_HOSTDEVICE inline Dual<T,P> fabs (const Dual<T,P> &x)
 {
     return x.val() >= T(0) ? x : -x;
 }
 
 
 
-inline float smoothstep(float e0, float e1, float x) { 
+OSL_HOSTDEVICE inline float smoothstep(float e0, float e1, float x) { 
     if (x < e0) return 0.0f;
     else if (x >= e1) return 1.0f;
     else {
@@ -997,7 +1005,7 @@ inline float smoothstep(float e0, float e1, float x) {
 
 // f(t) = (3-2t)t^2,   t = (x-e0)/(e1-e0)
 template<class T, int P>
-inline Dual<T,P> smoothstep (const Dual<T,P> &e0, const Dual<T,P> &e1, const Dual<T,P> &x)
+OSL_HOSTDEVICE inline Dual<T,P> smoothstep (const Dual<T,P> &e0, const Dual<T,P> &e1, const Dual<T,P> &x)
 {
    if (x.val() < e0.val()) {
       return Dual<T,P> (T(0));
@@ -1013,7 +1021,7 @@ inline Dual<T,P> smoothstep (const Dual<T,P> &e0, const Dual<T,P> &e1, const Dua
 
 // ceil(Dual) loses derivatives
 template<class T, int P>
-inline constexpr float ceil (const Dual<T,P> &x)
+OSL_HOSTDEVICE inline constexpr float ceil (const Dual<T,P> &x)
 {
     return std::ceil(x.val());
 }
@@ -1021,7 +1029,7 @@ inline constexpr float ceil (const Dual<T,P> &x)
 
 // floor(Dual) loses derivatives
 template<class T, int P>
-inline constexpr float floor (const Dual<T,P> &x)
+OSL_HOSTDEVICE inline constexpr float floor (const Dual<T,P> &x)
 {
     return std::floor(x.val());
 }
@@ -1029,7 +1037,7 @@ inline constexpr float floor (const Dual<T,P> &x)
 
 // floor, cast to an int.
 template<class T, int P>
-inline constexpr int
+OSL_HOSTDEVICE inline constexpr int
 ifloor (const Dual<T,P> &x)
 {
     return (int)floor(x);
